%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% MATLAB code
% File: ChromosomeCoding.m
%
% Purpose: Main file for running the Genetic Algorithm
%          contains coding of the chromosome.
%
% Reference:
% Battaglia, F. and Protopapas, M.K. (2011).
%   Time-varying multi-regime models fitting by genetic algorithms.
%   Journal of Time Series Analysis, 32, 237-252.
%   DOI: 10.1111/j.1467-9892.2010.00695.x.
%
% A related paper:
% Lundbergh, S., Ter\"{a}svirta, T., and van Dijk, D. (2003).
%   Time-varying smooth transition autoregressive models.
%   Journal of Business & Economic Statistics, 21(1), 104-121.
%   DOI: 10.1198/073500102288618810.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% Quantities to be chosen
populationSize      = 50 ; %number of chromosomes in population
probabilityMutation = 0.05; 
numberIterations    = 200;   %number of generations
datafiletoberead='...................'; % datafile, ascii code, one observation each row
%
%
% open file for ready
FID            = fopen(datafiletoberead, 'r');
timeSeriesData = fscanf(FID, '%g' , [1  inf]);    % It has two rows now.
timeSeriesData = timeSeriesData';
status         = fclose(FID);
%
% Start of GA **********************
STAR_MODEL         = 1;
PLTAR_MODEL        = 2;
GENERAL_STAR_MODEL = 3;
STATIONARY_MODEL   = 4;
numberObservations = length(timeSeriesData);
% bits 1,2 modelInTime 00,11 =stat, 01=STAR, 10=PLTAR
% bits 3,4 modelInDelay same as b4

minimumObservationsPerDomain = floor(numberObservations*.2);

% threshold in time should be between minimumObservationsPerDomain+1 (coded
% as 0) and numberObservations-*minimumObservationsPerDomain (coded as
% numberObservations-2*minimumObservationsPerDomain, so
displacementTypeTime       = 1;
displacementTypeDelay      = 3;
displacementThresholdTime  = 5;
bitsThresholdTime          = 12;
displacementThresholdDelay = displacementThresholdTime+bitsThresholdTime;

% threshold in delay should be between min, max of timeSeriesData for times
% t1 = minimumObservationsPerDomain and
% t2 = numberObservations-minimumObservationsPerDomain
sortedData = sort(timeSeriesData); % to get the threshold

% for i=1:numberObservations-2*minimumObservationsPerDomain
%    truncatedSortedData(i) = sortedData(i+minimumObservationsPerDomain);
% end

bits4Order         = 3; 
bits4Delay         = 3; 
bitsThresholdDelay = 12;
displacementOrder  = displacementThresholdDelay+bitsThresholdDelay;

% order of the models between 0 and 7, say
displacementDelay = displacementOrder+bits4Order;

% delays between 1 and 8, say
displacementGammaTime = displacementDelay+bits4Delay;
minGammaTime          = 46/numberObservations;
maxGammaTime          = 4.6;
baseTime              = 1.1;
numberBits4GammaTime  = 7;

% exponential transformation to have more chromosome values towards the min
% value
% realGamma=
% ((maxGammaTime-minGammaTime)/(baseTime^(2^numberBits4GammaTime-1)-1))*
% baseTime^bitValueGammaTime+minGammaTime-
% ((maxGammaTime-minGammaTime)/(baseTime^(2^numberBits4GammaTime-1)-1))

displacementGammaDelay = displacementGammaTime+numberBits4GammaTime;
minGammaDelay          = 46/(sortedData(numberObservations)-sortedData(1));
maxGammaDelay          = 4.6*numberObservations/(sortedData(numberObservations)-sortedData(1));
baseDelay              = 2;
numberBits4GammaDelay  = 7;
vectorDisplacements    = [displacementTypeTime displacementTypeDelay displacementThresholdTime displacementThresholdDelay displacementOrder displacementDelay displacementGammaTime displacementGammaDelay];

% finally we can calulate the chromosome's length)
chromosomeLength = displacementGammaDelay+numberBits4GammaDelay;
chromosomeSize   = chromosomeLength;

% bestAICSoFar      = +inf;
% blankObservations = 0; % initial and final observations cannot be checked for
% outliers due to lack of data

bestChromosomeSoFar = zeros(1,chromosomeSize);
bestFitnessSoFar    = -exp(1000);
populationFitness   = zeros(1,populationSize);
childrenMatrix      = zeros(populationSize,chromosomeSize);
cumulativeFitness   = zeros(1,populationSize);
populationMatrix    = zeros(populationSize,chromosomeSize);

for i=1:populationSize      % random initial population
    for j=1:chromosomeSize
        if rand<.5
            populationMatrix(i,j) = 1;
        else 
            populationMatrix(i,j) = 0;
        end
    end
end

initPop = populationMatrix;
for generationNumber=1:numberIterations % Now the main algorithm starts,
    generationNumber
    % and it executes for the given time of
    % iterations, stored in numberIterations
    
    % decode chrom
    for i0=1:populationSize
        for j=1:chromosomeSize
            chrom(j) = populationMatrix(i0,j);
        end
        if  chrom(1)==1 && chrom(2)==0
            typeOfModelinTime=STAR_MODEL;
        elseif chrom(1)==0 && chrom(2)==1
            typeOfModelinTime = PLTAR_MODEL;
        else
            typeOfModelinTime = STATIONARY_MODEL;
        end
        
        % modelInDelay
        if  chrom(3)==1 && chrom(4)==0
            typeOfModelInDelay=STAR_MODEL;
        elseif chrom(3)==0 && chrom(4)==1
            typeOfModelInDelay = PLTAR_MODEL;
        else
            typeOfModelInDelay = STATIONARY_MODEL;
        end
        
        % orderOfModels
        orderOfModels = 0;
        for i=1:bits4Order
            orderOfModels = orderOfModels+chrom(displacementOrder+i-1)*2^(i-1);
        end
        
        % lagDelay
        lagDelay = 1;
        for i=1:bits4Delay
            lagDelay = lagDelay+chrom(displacementDelay+i-1)*2^(i-1);
        end
        
        % ThresholdTime
        bitThreshold = 0;
        for i=1:bitsThresholdTime
            bitThreshold = bitThreshold+chrom(displacementThresholdTime+i-1)*2^(i-1);
        end
        thresholdTime = round(bitThreshold*(numberObservations-2*minimumObservationsPerDomain)/(2^bitsThresholdTime-1))+minimumObservationsPerDomain;

        % ThresholdDelay
        bitThreshold = 0;
        for i=1:bitsThresholdDelay;
            bitThreshold = bitThreshold+chrom(displacementThresholdDelay+i-1)*2^(i-1);
        end
        thresholdDelay = round(bitThreshold*(numberObservations-2*minimumObservationsPerDomain)/(2^bitsThresholdDelay-1))+minimumObservationsPerDomain;

        % finally get appropriate value from sorted vecotr
        thresholdDelay = sortedData(thresholdDelay);
        % GammaTime
        bitThreshold = 0;
        for i=1:numberBits4GammaTime
            bitThreshold = bitThreshold+chrom(displacementGammaTime+i-1)*2^(i-1);
        end
        
        % exponential transformation to have more chromosome values towards the min
        % value
        gammaCoefficientTime = ((maxGammaTime-minGammaTime)/(baseTime^(2^numberBits4GammaTime-1)-1))*(baseTime^bitThreshold-1)+minGammaTime;

        % gammaDelay
        bitThreshold = 0;
        for i=1:numberBits4GammaDelay
            bitThreshold = bitThreshold+chrom(displacementGammaDelay+i-1)*2^(i-1);
        end
        
        % exponential transformation to have more chromosome values towards the min
        % value
        gammaCoefficientDelay = ((maxGammaDelay-minGammaDelay)/(baseDelay^(2^numberBits4GammaDelay-1)-1))*(baseDelay^bitThreshold-1)+minGammaDelay;
    
        populationFitness(i0) = EstimateCoefficients(timeSeriesData,typeOfModelinTime,typeOfModelInDelay, orderOfModels,lagDelay,thresholdTime,thresholdDelay,gammaCoefficientTime,gammaCoefficientDelay);
        if populationFitness(i0)>bestFitnessSoFar  % update best chromosome
            bestFitnessSoFar    = populationFitness(i0);
            bestChromosomeSoFar = chrom;
        end
    end % for populationSize
     
    %%% The pairings of parents are based on the roulette rule.
    for s=1:populationSize/2          
        cumulativeFitness = cumsum(populationFitness);
        c                 = max(cumulativeFitness);
        
        % selection
        % picking first parent
        x       = c*rand;   % x is a random number between 0 and total cumulative
        d       = min(find(cumulativeFitness>x));
        parent1 = populationMatrix(d,:);
        
        % picking second parent
        x       = c*rand;   %x is a random number between 0 and total cumulative
        d       = min(find(cumulativeFitness>x));
        parent2 = populationMatrix(d,:);
        
       % random cutting point, use entire bit-strings
       c   = ceil(rand(1)*7);
       cut = vectorDisplacements(c);
       for i=1:cut-1
           child1(i) = parent1(i);
           child2(i) = parent2(i);
       end
       
       for i=cut:chromosomeLength
           child1(i) = parent2(i);
           child2(i) = parent1(i);
       end 
       
       % mutation
        mask   = rand(1,chromosomeSize)<probabilityMutation;
        child1 = xor(child1,mask);
        mask   = rand(1,chromosomeSize)<probabilityMutation;
        child2 = xor(child2,mask);
        
        % store children in children matrix
        p1 = fix(2*s-1);  % since a pairing produces 2 children
        p2 = 2*s;
        % for j=1:chromosomeSize
        %    childrenMatrix(p1,j) = child1(j);
        %    childrenMatrix(p2,j) = child2(j);
        % end
        childrenMatrix(p1,:) = child1;
        childrenMatrix(p2,:) = child2;
    end %pairing loop for s=1:fix(populationSize/2)
% (rough) elitism
childrenMatrix(1,:) = bestChromosomeSoFar;
%
    populationMatrix = childrenMatrix; % no tournament  
    history(generationNumber)   = mean(populationFitness);
    historfit(generationNumber) = bestFitnessSoFar;
end  % generations
%
%
% decode best Chromosome and output results
%
%
    for j=1:chromosomeSize
        chrom(j) = bestChromosomeSoFar(j);
    end
    if  chrom(1)==1 && chrom(2)==0
        typeOfModelinTime = STAR_MODEL;
    elseif chrom(1)==0 && chrom(2)==1
        typeOfModelinTime = PLTAR_MODEL;
    else
        typeOfModelinTime = STATIONARY_MODEL;
    end
    
    % modelInDelay
    if  chrom(3)==1 && chrom(4)==0
        typeOfModelInDelay = STAR_MODEL;
    elseif chrom(3)==0 && chrom(4)==1
        typeOfModelInDelay = PLTAR_MODEL;
    else
        typeOfModelInDelay = STATIONARY_MODEL;
    end
    
    % orderOfModels
    orderOfModels = 0;
    for i=1:bits4Order
        orderOfModels = orderOfModels+chrom(displacementOrder+i-1)*2^(i-1);
    end
    
    % lagDelay
    lagDelay = 1;
    for i=1:bits4Delay
        lagDelay = lagDelay+chrom(displacementDelay+i-1)*2^(i-1);
    end
    
    % ThresholdTime
    bitThreshold = 0;
    for i=1:bitsThresholdTime
        bitThreshold = bitThreshold+chrom(displacementThresholdTime+i-1)*2^(i-1);
    end
    
    thresholdTime = round(bitThreshold*(numberObservations-2*minimumObservationsPerDomain)/(2^bitsThresholdTime-1))+minimumObservationsPerDomain;

    % ThresholdDelay
    bitThreshold = 0;
    for i=1:bitsThresholdDelay;
        bitThreshold = bitThreshold+chrom(displacementThresholdDelay+i-1)*2^(i-1);
    end
    
    thresholdDelay = round(bitThreshold*(numberObservations-2*minimumObservationsPerDomain)/(2^bitsThresholdDelay-1))+minimumObservationsPerDomain;

    % finally get appropriate value from sorted vecotr
    thresholdDelay = sortedData(thresholdDelay);
    % GammaTime
    bitThreshold = 0;
    for i=1:numberBits4GammaTime
        bitThreshold = bitThreshold+chrom(displacementGammaTime+i-1)*2^(i-1);
    end
    
    % exponential transformation to have more chromosome values towards the min
    % value
    gammaCoefficientTime = ((maxGammaTime-minGammaTime)/(baseTime^(2^numberBits4GammaTime-1)-1))*baseTime^bitThreshold+minGammaTime-((maxGammaTime-minGammaTime)/(baseTime^(2^numberBits4GammaTime-1)-1));

    % gammaDelay
    bitThreshold = 0;
    for i=1:numberBits4GammaDelay
        bitThreshold = bitThreshold+chrom(displacementGammaDelay+i-1)*2^(i-1);
    end
    
    % exponential transformation to have more chromosome values towards the min
    % value
    gammaCoefficientDelay = ((maxGammaDelay-minGammaDelay)/(baseDelay^(2^numberBits4GammaDelay-1)-1))*baseDelay^bitThreshold+minGammaDelay-((maxGammaDelay-minGammaDelay)/(baseDelay^(2^numberBits4GammaDelay-1)-1));
    [fitness,estimatedCoefficients,squareSumRes,residualVariance,residuals] = EstimateCoefficients (timeSeriesData, typeOfModelinTime, typeOfModelInDelay, orderOfModels,lagDelay,thresholdTime,thresholdDelay,gammaCoefficientTime,gammaCoefficientDelay);
    historfit
    typeOfModelinTime
    typeOfModelInDelay
    orderOfModels
    lagDelay
    thresholdTime
    thresholdDelay
    fitness