%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% MATLAB code
% File: EstimateCoefficients.m
%
% Genetic Algorithm for multi-regime nonlinear nonstationary project.
%
% Purpose: To estimate the coefficients, given time series data, and other
%          parameters, to produce the fitness function for the GA.
%
% Reference:
% Battaglia, F. and Protopapas, M.K. (2011).
%   Time-varying multi-regime models fitting by genetic algorithms.
%   Journal of Time Series Analysis, 32, 237-252.
%   DOI: 10.1111/j.1467-9892.2010.00695.x.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

function [fitness,estimatedCoefficients,squareSumRes,residualVariance,residuals] = EstimateCoefficients(timeSeriesData,typeOfModelinTime,typeOfModelInDelay,orderOfModels,lagDelay,thresholdTime,thresholdDelay,gammaCoefficientTime,gammaCoefficientDelay)
STAR_MODEL         = 1;                     
PLTAR_MODEL        = 2;
GENERAL_STAR_MODEL = 3;
STATIONARY_MODEL   = 4;
%
% choose the value of the constant c in penalized likelihood
%
  AICConstant = 3;
%
%
%estimatedCoef = [a,b,c,d,f,g,h,k,m];
%coef0         = [1 t H_t y(t-d) ty(t-d) y(t-d)H_t H tH H_tH];
%coefj         = y(t-j)*coef0;

estimatedCoefficients = zeros(1,(orderOfModels+1)*9);
startingIndex         = max(orderOfModels,lagDelay);
p                     = startingIndex;
numberObservations    = length(timeSeriesData);
y                     = timeSeriesData(startingIndex+1:numberObservations);

if (typeOfModelinTime==STATIONARY_MODEL) && (typeOfModelInDelay==STATIONARY_MODEL)  
    for i=1:numberObservations-startingIndex
        for j=2:orderOfModels+1
            X(i,j) = timeSeriesData(i-j+p+1);
        end
    X(i,1) = 1;
    end
    betas = (X'*X)^(-1)*X'*y;
    for i=1:orderOfModels+1
        estimatedCoefficients((i-1)*9+1) = betas(i);
    end
elseif (typeOfModelinTime==STATIONARY_MODEL) && (typeOfModelInDelay==STAR_MODEL)
     % all linear and functional terms in time are zero
     % linear terms on the lags (1,..,p) are zero
     for i=1:numberObservations-startingIndex
         for j=2:orderOfModels+1
             X(i,j)                 = timeSeriesData(i-j+p+1);
             X(i,j+orderOfModels+1) = timeSeriesData(i-j+p+1)/(1+exp(-gammaCoefficientDelay*(timeSeriesData(i+p-lagDelay)-thresholdDelay)));
         end
     X(i,1) = 1;
     end
     for i=1:numberObservations-startingIndex
         X(i,orderOfModels+2) = 1/(1+exp(-gammaCoefficientDelay*(timeSeriesData(i+p-lagDelay)-thresholdDelay)));
     end
     betas = (X'*X)^(-1)*X'*y;
     for i=1:orderOfModels+1
         estimatedCoefficients((i-1)*9+1) = betas(i);
         estimatedCoefficients((i-1)*9+7) = betas(i+orderOfModels+1);
     end     
    elseif (typeOfModelinTime==STATIONARY_MODEL) && (typeOfModelInDelay==PLTAR_MODEL)
         % stationary in time means b_j=c_j=f_j=k_j=m_j=0    ,j=1,...
         % pltar in levels implies h_0=k_0=m_0=0
         % we have a0,...ap,d1,...,dp,h1,...,hp
     for i=1:numberObservations-startingIndex
         for j=2:orderOfModels+1
             X(i,j)                   = timeSeriesData(i-j+p+1);
             X(i,j+orderOfModels)     = timeSeriesData(i-j+p+1)*timeSeriesData(i+p-lagDelay);
             X(i,j+2*orderOfModels+1) = timeSeriesData(i-j+p+1)*max([0 timeSeriesData(i+p-lagDelay)-thresholdDelay]);
         end
     X(i,1) = 1;
     end    
     for i=1:numberObservations-startingIndex
         X(i,2*orderOfModels+2) = max([0 timeSeriesData(i+p-lagDelay)-thresholdDelay]);
     end    
     betas = (X'*X)^(-1)*X'*y;
     for i=1:orderOfModels+1
         estimatedCoefficients((i-1)*9+1) = betas(i);
         if i<=orderOfModels
            estimatedCoefficients(i*9+4)  = betas(i+orderOfModels+1);
         end
         estimatedCoefficients((i-1)*9+7) = betas(i+2*orderOfModels+1);
     end
     elseif (typeOfModelinTime==STAR_MODEL) && (typeOfModelInDelay==STATIONARY_MODEL) 
     for i=1:numberObservations-startingIndex
       for j=2:orderOfModels+1
           X(i,j)                 = timeSeriesData(i-j+p+1);
           X(i,j+orderOfModels+1) = timeSeriesData(i-j+p+1)/(1+exp(-gammaCoefficientTime*(i+p-thresholdTime)));
       end
     X(i,1) = 1;
     end
     for i=1:numberObservations-startingIndex
        X(i,orderOfModels+2) = 1/(1+exp(-gammaCoefficientTime*(i+p-thresholdTime)));
     end
     betas = (X'*X)^(-1)*X'*y;
     for i=1:orderOfModels+1
         estimatedCoefficients((i-1)*9+1) = betas(i);
         estimatedCoefficients((i-1)*9+3) = betas(i+orderOfModels+1);
     end
     elseif (typeOfModelinTime==PLTAR_MODEL) && (typeOfModelInDelay==STATIONARY_MODEL)         
     for i=1:numberObservations-startingIndex
         for j=2:orderOfModels+1
             X(i,j)                   = timeSeriesData(i-j+p+1);
             X(i,j+orderOfModels+1)   = timeSeriesData(i-j+p+1)*(i+p);
             X(i,j+2*orderOfModels+2) = timeSeriesData(i-j+p+1)*max([0 i+p-thresholdTime]);
         end
     X(i,1) = 1;
     end    
     for i=1:numberObservations-startingIndex
         X(i,2+orderOfModels)   = i+p;
         X(i,2*orderOfModels+3) = max([0 i+p-thresholdTime]);
     end    
     betas = (X'*X)^(-1)*X'*y;
     for i=1:orderOfModels+1
         estimatedCoefficients((i-1)*9+1) = betas(i);
         estimatedCoefficients((i-1)*9+2) = betas(i+orderOfModels+1);
         estimatedCoefficients((i-1)*9+3) = betas(i+2*orderOfModels+2);
     end
      elseif (typeOfModelinTime==STAR_MODEL) && (typeOfModelInDelay==STAR_MODEL) 
            for i=1:numberObservations-startingIndex
                for j=2:orderOfModels+1
                    X(i,j)                   = timeSeriesData(i-j+p+1);
                    X(i,j+orderOfModels+1)   = timeSeriesData(i-j+p+1)/(1+exp(-gammaCoefficientTime*(i+p-thresholdTime)));
                    X(i,j+2*orderOfModels+2) = timeSeriesData(i-j+p+1)/(1+exp(-gammaCoefficientDelay*(timeSeriesData(i+p-lagDelay)-thresholdDelay)));
                    X(i,j+3*orderOfModels+3) = X(i,j+orderOfModels+1)/(1+exp(-gammaCoefficientDelay*(timeSeriesData(i+p-lagDelay)-thresholdDelay)));
                end
            X(i,1)=1;
            end
    for i=1:numberObservations-startingIndex
        X(i,orderOfModels+2)   = 1/(1+exp(-gammaCoefficientTime*(i+p-thresholdTime)));
        X(i,2*orderOfModels+3) = 1/(1+exp(-gammaCoefficientDelay*(timeSeriesData(i+p-lagDelay)-thresholdDelay)));   
        X(i,3*orderOfModels+4) = 1/((1+exp(-gammaCoefficientDelay*(timeSeriesData(i+p-lagDelay)-thresholdDelay)))*(1+exp(-gammaCoefficientTime*(i+p-thresholdTime))));
    end
    betas = (X'*X)^(-1)*X'*y;
    for i=1:orderOfModels+1
        estimatedCoefficients((i-1)*9+1) = betas(i);
        estimatedCoefficients((i-1)*9+3) = betas(i+orderOfModels+1);
        estimatedCoefficients((i-1)*9+7) = betas(i+2*orderOfModels+2);
        estimatedCoefficients((i-1)*9+9) = betas(i+3*orderOfModels+3);      
    end
      elseif (typeOfModelinTime==STAR_MODEL) && (typeOfModelInDelay==PLTAR_MODEL) 
            for i=1:numberObservations-startingIndex
                for j=2:orderOfModels+1
                    X(i,j)                   = timeSeriesData(i-j+p+1);%a
                    X(i,j+orderOfModels+1)   = timeSeriesData(i-j+p+1)/(1+exp(-gammaCoefficientTime*(i+p-thresholdTime)));%c
                    X(i,j+2*orderOfModels+1) = timeSeriesData(i-j+p+1)*timeSeriesData(i+p-lagDelay);%d                 
                    X(i,j+3*orderOfModels+1) = timeSeriesData(i-j+p+1)*timeSeriesData(i+p-lagDelay)/(1+exp(-gammaCoefficientTime*(i+p-thresholdTime)));
                    X(i,j+4*orderOfModels+2) = timeSeriesData(i-j+p+1)*max([0 timeSeriesData(i+p-lagDelay)-thresholdDelay]);
                    X(i,j+5*orderOfModels+3) = timeSeriesData(i-j+p+1)*max([0 timeSeriesData(i+p-lagDelay)-thresholdDelay])/(1+exp(-gammaCoefficientTime*(i+p-thresholdTime)));
                end
                 X(i,1) = 1;
           end

    for i=1:numberObservations-startingIndex
        X(i,orderOfModels+2)   = 1/(1+exp(-gammaCoefficientTime*(i+p-thresholdTime)));
        X(i,4*orderOfModels+3) = max([0 timeSeriesData(i+p-lagDelay)-thresholdDelay]);
        X(i,5*orderOfModels+4) = max([0 timeSeriesData(i+p-lagDelay)-thresholdDelay])/(1+exp(-gammaCoefficientTime*(i+p-thresholdTime)));
    end
    betas = (X'*X)^(-1)*X'*y;
    for i=1:orderOfModels+1
        estimatedCoefficients((i-1)*9+1)  = betas(i);
        estimatedCoefficients((i-1)*9+3)  = betas(i+orderOfModels+1);
         estimatedCoefficients((i-1)*9+7) = betas(i+4*orderOfModels+2);
         estimatedCoefficients((i-1)*9+9) = betas(i+5*orderOfModels+3); 
         if i<=orderOfModels
            estimatedCoefficients(i*9+4)  = betas(i+2*orderOfModels+2);
            estimatedCoefficients(i*9+6)  = betas(i+3*orderOfModels+2);
        end  
    end
      elseif (typeOfModelinTime==PLTAR_MODEL) && (typeOfModelInDelay==STAR_MODEL) 
            for i=1:numberObservations-startingIndex
                for j=2:orderOfModels+1
                    X(i,j)                   = timeSeriesData(i-j+p+1);%a
                    X(i,j+orderOfModels+1)   = timeSeriesData(i-j+p+1)*(i+p);%b
                    X(i,j+2*orderOfModels+2) = timeSeriesData(i-j+p+1)*max([0 i+p-thresholdTime]);%c               
                    X(i,j+3*orderOfModels+3) = timeSeriesData(i-j+p+1)/(1+exp(-gammaCoefficientDelay*(timeSeriesData(i+p-lagDelay)-thresholdDelay)));%h
                    X(i,j+4*orderOfModels+4) = timeSeriesData(i-j+p+1)*(i+p)/(1+exp(-gammaCoefficientDelay*(timeSeriesData(i+p-lagDelay)-thresholdDelay))); %k
                    X(i,j+5*orderOfModels+5) = timeSeriesData(i-j+p+1)*max([0 i+p-thresholdTime])/(1+exp(-gammaCoefficientDelay*(timeSeriesData(i+p-lagDelay)-thresholdDelay)));%m
                end
            X(i,1) = 1;
            end
    for i=1:numberObservations-startingIndex
        X(i,orderOfModels+2)   = (i+p);
        X(i,2*orderOfModels+3) = max([0 i+p-thresholdTime]);
        X(i,3*orderOfModels+4) = 1/(1+exp(-gammaCoefficientDelay*(timeSeriesData(i+p-lagDelay)-thresholdDelay)));
        X(i,4*orderOfModels+5) = (i+p)/(1+exp(-gammaCoefficientDelay*(timeSeriesData(i+p-lagDelay)-thresholdDelay))); %k
        X(i,5*orderOfModels+6) = max([0 i+p-thresholdTime])/(1+exp(-gammaCoefficientDelay*(timeSeriesData(i+p-lagDelay)-thresholdDelay)));%m
    end
    betas = (X'*X)^(-1)*X'*y;
    for i=1:orderOfModels+1
        estimatedCoefficients((i-1)*9+1)  = betas(i);
        estimatedCoefficients((i-1)*9+2)  = betas(i+orderOfModels+1);
         estimatedCoefficients((i-1)*9+3) = betas(i+2*orderOfModels+2);
         estimatedCoefficients((i-1)*9+7) = betas(i+3*orderOfModels+3); 
         estimatedCoefficients((i-1)*9+8) = betas(i+4*orderOfModels+4); 
         estimatedCoefficients((i-1)*9+9) = betas(i+5*orderOfModels+5); 
    end
        %%%%%new
        elseif (typeOfModelinTime==PLTAR_MODEL) && (typeOfModelInDelay==PLTAR_MODEL) 
            for i=1:numberObservations-startingIndex
                for j=2:orderOfModels+1
                    X(i,j)                   = timeSeriesData(i-j+p+1);%a
                    X(i,j+orderOfModels+1)   = timeSeriesData(i-j+p+1)*(i+p);%b
                    X(i,j+2*orderOfModels+2) = timeSeriesData(i-j+p+1)*max([0 i+p-thresholdTime]);%c    
                    X(i,j+3*orderOfModels+2) = timeSeriesData(i-j+p+1)*timeSeriesData(i+p-lagDelay);%d  
                    X(i,j+4*orderOfModels+2) = timeSeriesData(i-j+p+1)*(i+p)*timeSeriesData(i+p-lagDelay);%f
                    X(i,j+5*orderOfModels+2) = timeSeriesData(i-j+p+1)*max([0 i+p-thresholdTime])*timeSeriesData(i+p-lagDelay);%g
                    X(i,j+6*orderOfModels+3) = timeSeriesData(i-j+p+1)*max([0 timeSeriesData(i+p-lagDelay)-thresholdDelay]);  %h    
                    X(i,j+7*orderOfModels+4) = timeSeriesData(i-j+p+1)*(i+p)*max([0 timeSeriesData(i+p-lagDelay)-thresholdDelay]); %k
                    X(i,j+8*orderOfModels+5) = timeSeriesData(i-j+p+1)*max([0 i+p-thresholdTime])*max([0 timeSeriesData(i+p-lagDelay)-thresholdDelay]); %m
                end
            X(i,1) = 1;
            end
        for i=1:numberObservations-startingIndex
            X(i,orderOfModels+2)   = (i+p);%b
            X(i,2*orderOfModels+3) = max([0 i+p-thresholdTime]);%c
            X(i,6*orderOfModels+4) = max([0 timeSeriesData(i+p-lagDelay)-thresholdDelay]);%h
            X(i,7*orderOfModels+5) = (i+p)*max([0 timeSeriesData(i+p-lagDelay)-thresholdDelay]); %k
            X(i,8*orderOfModels+6) = max([0 i+p-thresholdTime])*max([0 timeSeriesData(i+p-lagDelay)-thresholdDelay]); %m
         
        end
%%      xm    = X'*X;
        betas = (X'*X)^(-1)*X'*y;
        for i=1:orderOfModels+1
            estimatedCoefficients((i-1)*9+1) = betas(i);
            estimatedCoefficients((i-1)*9+2) = betas(i+orderOfModels+1);
            estimatedCoefficients((i-1)*9+3) = betas(i+2*orderOfModels+2);
            estimatedCoefficients((i-1)*9+7) = betas(i+6*orderOfModels+3); 
            estimatedCoefficients((i-1)*9+8) = betas(i+7*orderOfModels+4);
            estimatedCoefficients((i-1)*9+9) = betas(i+8*orderOfModels+5); 
            if i<=orderOfModels
                estimatedCoefficients(i*9+4) = betas(i+3*orderOfModels+3);
                estimatedCoefficients(i*9+5) = betas(i+4*orderOfModels+3);
                estimatedCoefficients(i*9+6) = betas(i+5*orderOfModels+3);
            end 
        end
end       
      
     
  % finally calculate residuals and their sum of squares
  residuals          = y-X*betas;
  squareSumRes       = residuals'*residuals; 
  residualVariance   = squareSumRes/(numberObservations-startingIndex);
  numberParameters   = length(betas);
  numberObservations = length(y);
  
  fitness            = exp(-AICConstant*numberParameters/numberObservations)/residualVariance;
  
    
